import tensorflow as tf
from transform import feature_spec, postprocess, create_tokenizer
import sys, os

flags = tf.flags
FLAGS = flags.FLAGS

flags.DEFINE_string("assets_dir", "assets", "The assets directory generated by assets.py.")
flags.DEFINE_string("tfrecords_file", "dataset/test_128.tfrecords", "The tfrecords file to be inspected.")
flags.DEFINE_integer("max_seq_length", 128, "Maximum sequence length.")
flags.DEFINE_integer("display_rows", 1, "Number of rows to be displayed.")

def view_tfrecord(file_name, feature_spec, display_rows=5):
    """
    >>> view_tfrecord('./dataset/val_128.tfrecords', 128)
    """
    tokenizer = create_tokenizer(
        os.path.join(FLAGS.assets_dir, 'chinese_L-12_H-768_A-12', 'vocab.txt')
    )
    with tf.Session() as sess:
        i = 0
        for example in tf.python_io.tf_record_iterator(file_name):
            if i >= display_rows:
                break
            i+=1
            features = tf.parse_single_example(example, features=feature_spec)
            print('Example %d:'%i)
            for feature_name, tensor in features.items():
                if feature_name == 'text':
                    print(feature_name, features[feature_name].eval().values[0].decode('utf-8'))
                else:
                    print(feature_name, features[feature_name].eval())

            text = features['text'].eval().values[0].decode('utf-8')
            bert_tokens_len = features['bert_tokens_len'].eval()[0]
            bert_tokens = tokenizer.convert_ids_to_tokens(features['input_ids'].eval()[1:bert_tokens_len+1])
            bert_truths = features['truths'].eval()[1:bert_tokens_len+1]

            print("Final Result:", postprocess(text, bert_tokens, bert_truths, 0.5))



def main(_):
    tf.logging.set_verbosity(tf.logging.INFO)

    sys.path += [os.path.join(FLAGS.assets_dir, 'bert')]
    view_tfrecord(
        FLAGS.tfrecords_file,
        feature_spec(FLAGS.max_seq_length),
        FLAGS.display_rows
    )


if __name__ == "__main__":
  tf.app.run()



import os
import tensorflow as tf
import pandas as pd
from tqdm import tqdm
from sklearn.model_selection import train_test_split
import sys

flags = tf.flags
FLAGS = flags.FLAGS

flags.DEFINE_string("assets_dir", "assets", "The assets directory generated by assets.py.")
flags.DEFINE_string("output_dir", "dataset", "The output directory.")
flags.DEFINE_integer("max_seq_length", 512, "Maximum sequence length.")
flags.DEFINE_integer("random_seed", 42, "Random seed.")

def main(_):
    tf.logging.set_verbosity(tf.logging.INFO)
    tf.gfile.MakeDirs(FLAGS.output_dir)

    sys.path += [os.path.join(FLAGS.assets_dir, 'bert')]
    from transform import create_tokenizer, text_to_tfexample

    tokenizer = create_tokenizer(
        os.path.join(FLAGS.assets_dir, 'chinese_L-12_H-768_A-12', 'vocab.txt')
    )

    config = {}
    for dataset in ['pku', 'msr', 'cityu', 'as']:
        print(os.path.join(FLAGS.assets_dir, 'icwb2-data', 'training', '%s_training.utf8'%dataset))
        config['train_%s'%dataset] = pd.read_csv(
            os.path.join(FLAGS.assets_dir, 'icwb2-data', 'training', '%s_training.utf8'%dataset),
            header=None, sep='^'
        )[0]
        print(os.path.join(FLAGS.assets_dir, 'icwb2-data', 'testing', '%s_test.utf8'%dataset))
        config['test_%s'%dataset] = pd.read_csv(
            os.path.join(FLAGS.assets_dir, 'icwb2-data', 'testing', '%s_test.utf8'%dataset),
            header=None, sep='^'
        )[0]

    for type, ser in config.items():
        outfile = os.path.join(FLAGS.output_dir, '%s_%s.tfrecords'%(type, FLAGS.max_seq_length))
        with tf.python_io.TFRecordWriter(outfile) as writer:
            tf.logging.info('Writing to %s...'%outfile)
            for text in tqdm(ser):
                writer.write(text_to_tfexample(text, FLAGS.max_seq_length, tokenizer).SerializeToString())


if __name__ == "__main__":
  tf.app.run()



import os
import ntpath
import requests
from io import BytesIO
from zipfile import ZipFile
from pathlib import Path
import requests
from tqdm import tqdm
import math


def maybe_download_unzip(url, dst_path, verbose=False, force=False):
    dir = os.path.join(dst_path, os.path.splitext(ntpath.basename(url))[0])
    if os.path.exists(dir) and not force:
        if verbose:
            print('Directory %s already exists, skipping download...'%dir)
        return
    if verbose:
        print('Downloading %s to %s ...'%(url, dir))

    r = requests.get(url, stream=True)
    total_size_mb = math.ceil(int(r.headers.get('content-length', 0)) / 1024 / 1024);

    f = BytesIO()
    func = lambda x: tqdm(x, total=total_size_mb, unit='MB', unit_scale=True) if verbose else lambda x:x
    for chunk in func(r.iter_content(1024 * 1024)):
        f.write(chunk)
    ZipFile(f).extractall(dst_path)
    if verbose:
        print('Done downloading %s to %s ...'%(url, dir))
    return



from io import BytesIO
from zipfile import ZipFile
import tempfile
import requests
import subprocess
import os
import ntpath
import tensorflow as tf

flags = tf.flags
FLAGS = flags.FLAGS

flags.DEFINE_string("output_dir", "assets", "The output directory.")

def maybe_download_unzip(url, dst_path):
    dir = os.path.join(dst_path, os.path.splitext(ntpath.basename(url))[0])
    if os.path.exists(dir):
        tf.logging.info('Directory %s already exists, skipping download...'%dir)
        return
    tf.logging.info('Downloading %s...'%url)
    req = requests.get(url)
    zipfile = ZipFile(BytesIO(req.content))
    zipfile.extractall(dst_path)
    return

def maybe_git_clone(url, dst_path):
    repo_name = os.path.splitext(ntpath.basename(url))[0]
    dir = os.path.join(dst_path, repo_name)
    if os.path.exists(dir):
        tf.logging.info('Directory %s already exists, skipping download...'%dir)
        return

    tf.logging.info('Cloning %s...'%url)
    subprocess.call(['git', 'clone',
        url,
        os.path.join(dst_path, repo_name)
    ])
    return


def main(_):
    tf.logging.set_verbosity(tf.logging.INFO)
    tf.gfile.MakeDirs(FLAGS.output_dir)


    maybe_download_unzip(
        'http://sighan.cs.uchicago.edu/bakeoff2005/data/icwb2-data.zip',
        FLAGS.output_dir
    )

    maybe_download_unzip(
        'https://storage.googleapis.com/bert_models/2018_11_03/chinese_L-12_H-768_A-12.zip',
        FLAGS.output_dir
    )

    maybe_git_clone(
        'https://github.com/google-research/bert',
        FLAGS.output_dir
    )


if __name__ == "__main__":
  flags.mark_flag_as_required("output_dir")
  tf.app.run()



import modeling, optimization
import tensorflow as tf

def create_model(bert_config, is_training, input_ids, input_mask, segment_ids,
                 use_one_hot_embeddings):
  model = modeling.BertModel(
      config=bert_config,
      is_training=is_training,
      input_ids=input_ids,
      input_mask=input_mask,
      token_type_ids=segment_ids,
      use_one_hot_embeddings=use_one_hot_embeddings)

  final_hidden = model.get_sequence_output()

  final_hidden_shape = modeling.get_shape_list(final_hidden, expected_rank=3)
  batch_size = final_hidden_shape[0]
  seq_length = final_hidden_shape[1]
  hidden_size = final_hidden_shape[2]

  output_weights = tf.get_variable(
      "berserker/output_weights", [1, hidden_size],
      initializer=tf.truncated_normal_initializer(stddev=0.02))

  output_bias = tf.get_variable(
      "berserker/output_bias", [1], initializer=tf.zeros_initializer())

  final_hidden_matrix = tf.reshape(final_hidden,
                                   [batch_size * seq_length, hidden_size])

  logits = tf.matmul(final_hidden_matrix, output_weights, transpose_b=True)
  logits = tf.nn.bias_add(logits, output_bias)

  logits = tf.reshape(logits, [batch_size, seq_length])
  predictions = tf.nn.sigmoid(logits)

  return predictions


def model_fn_builder(bert_config, init_checkpoint, use_tpu, use_one_hot_embeddings,
                     learning_rate=None, num_train_steps=None, num_warmup_steps=None):
  """Returns `model_fn` closure for TPUEstimator."""

  def model_fn(features, labels, mode, params):  # pylint: disable=unused-argument
    """The `model_fn` for TPUEstimator."""

    tf.logging.info("*** Features ***")
    for name in sorted(features.keys()):
      tf.logging.info("  name = %s, shape = %s" % (name, features[name].shape))

    input_ids = features["input_ids"]
    input_mask = features["input_mask"]
    segment_ids = features["segment_ids"]
    truths = features["truths"]

    predictions = create_model(
        bert_config=bert_config,
        is_training=(mode == tf.estimator.ModeKeys.TRAIN),
        input_ids=input_ids,
        input_mask=input_mask,
        segment_ids=segment_ids,
        use_one_hot_embeddings=use_one_hot_embeddings)

    tvars = tf.trainable_variables()
    initialized_variable_names = {}

    #############
    # scaffold_fn
    #############
    scaffold_fn = None
    if init_checkpoint:
      (assignment_map, initialized_variable_names
      ) = modeling.get_assignment_map_from_checkpoint(tvars, init_checkpoint)
      if use_tpu:

        def tpu_scaffold():
          tf.train.init_from_checkpoint(init_checkpoint, assignment_map)
          return tf.train.Scaffold()

        scaffold_fn = tpu_scaffold
      else:
        tf.train.init_from_checkpoint(init_checkpoint, assignment_map)


    tf.logging.info("**** Trainable Variables ****")
    for var in tvars:
      init_string = ""
      if var.name in initialized_variable_names:
        init_string = ", *INIT_FROM_CKPT*"
      tf.logging.info("  name = %s, shape = %s%s", var.name, var.shape,
                      init_string)

    ############
    # total_loss
    ############
    if mode == tf.estimator.ModeKeys.TRAIN or mode == tf.estimator.ModeKeys.EVAL:
        input_mask_f = tf.cast(input_mask, tf.float32)
        per_example_loss = tf.reduce_sum(tf.keras.backend.binary_crossentropy(truths, predictions) * input_mask_f, axis=-1) / tf.reduce_sum(input_mask_f, axis=-1)
        total_loss = tf.reduce_mean(per_example_loss)


    #############
    # output_spec
    #############
    output_spec = None
    if mode == tf.estimator.ModeKeys.TRAIN:
      assert learning_rate is not None
      assert num_train_steps is not None
      assert num_warmup_steps is not None
      train_op = optimization.create_optimizer(
          total_loss,
          learning_rate,
          num_train_steps,
          num_warmup_steps,
          use_tpu)

      output_spec = tf.contrib.tpu.TPUEstimatorSpec(
          mode=mode,
          loss=total_loss,
          train_op=train_op,
          scaffold_fn=scaffold_fn)

    elif mode == tf.estimator.ModeKeys.EVAL:
      def metric_fn(truths, predictions, masks):
        threshold = 0.5
        pred_int = tf.cast(predictions >= threshold, tf.int32)
        # acc = tf.metrics.accuracy(labels=truths, predictions=predictions, weights=masks)
        return {
            "auc": tf.metrics.auc(
                labels=truths,
                predictions=predictions,
                weights=masks
            ),
            "f1_score @%.2f"%threshold: tf.contrib.metrics.f1_score(
                labels=truths,
                predictions=pred_int,
                weights=masks
            ),
            "precision @%.2f"%threshold: tf.metrics.precision(
                labels=truths,
                predictions=pred_int,
                weights=masks
            ),
            "recall @%.2f"%threshold: tf.metrics.recall(
                labels=truths,
                predictions=pred_int,
                weights=masks
            ),
            # "binary_crossentropy": total_loss
        }

      eval_metrics = (metric_fn, [truths, predictions, input_mask])
      output_spec = tf.contrib.tpu.TPUEstimatorSpec(
          mode=mode,
          loss=total_loss,
          eval_metrics=eval_metrics,
          scaffold_fn=scaffold_fn)
    else:
      output_spec = tf.contrib.tpu.TPUEstimatorSpec(
          mode=mode,
          predictions={
            "predictions": predictions
          },
          scaffold_fn=scaffold_fn)
    return output_spec

  return model_fn



from .utils import maybe_download_unzip
from .tokenization import FullTokenizer
from pathlib import Path
from .transform import text_to_bert_inputs, preprocess, postprocess
import tensorflow as tf
import numpy as np


_assets_path = Path(__file__).parent / 'assets'
_models_path = Path(__file__).parent / 'models'


MAX_SEQ_LENGTH = 512
SEQ_LENGTH = MAX_SEQ_LENGTH - 2


def load_model(model_name=None, verbose=True, force_download=False):
    maybe_download_unzip(
        'https://github.com/Hoiy/berserker/releases/download/v0.1-alpha/1547563491.zip',
        _models_path,
        verbose,
        force_download
    )


def tokenize(text):
    load_model()
    tokenizer = FullTokenizer(
        vocab_file=_assets_path / 'vocab.txt',
        do_lower_case=False
    )

    bert_inputs_lens = []
    bert_inputs = []
    temp = text
    while len(temp) > 0:
        bert_input = text_to_bert_inputs(temp[:SEQ_LENGTH], MAX_SEQ_LENGTH, tokenizer)
        bert_inputs_lens.append(len(preprocess(temp[:SEQ_LENGTH], tokenizer)[0]))
        bert_inputs.append(bert_input)
        temp = temp[SEQ_LENGTH:]

    berserker = tf.contrib.predictor.from_saved_model(
        str(_models_path / '1547563491')
    )
    output = berserker({
        'input_ids': [bi[0] for bi in bert_inputs],
        'input_mask': [bi[1] for bi in bert_inputs],
        'segment_ids': [bi[2] for bi in bert_inputs],
        'truths': [bi[3] for bi in bert_inputs]
    })

    results = output['predictions']

    results_itr = iter(results)
    bert_inputs_itr = iter(bert_inputs)
    bert_inputs_lens_itr = iter(bert_inputs_lens)

    prediction = np.array([])
    bert_tokens = []
    temp = text

    while len(temp) > 0:
        (input_ids, _, _, _) = next(bert_inputs_itr)
        bert_inputs_len = next(bert_inputs_lens_itr)
        result = next(results_itr)
        prediction = np.concatenate((prediction, result[1:1+bert_inputs_len]))
        bert_tokens += tokenizer.convert_ids_to_tokens(input_ids[1:1+bert_inputs_len])
        temp = temp[SEQ_LENGTH:]

    return postprocess(text, bert_tokens, prediction, threshold=0.5)



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name = 'basaka',
    version = '0.0.1',
    author = 'Hoi',
    author_email = 'hoiy927@gmail.com',
    description='Berserker - BERt chineSE tRained toKenizER',
    long_description=long_description,
    install_requires = ['requests', 'six', 'tqdm', 'tensorflow>=1.12.0'],
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)



import collections
import tensorflow as tf

def create_tokenizer(vocab_path):
    import tokenization
    return tokenization.FullTokenizer(
        vocab_file=vocab_path,
        do_lower_case=False
    )

def _to_text_and_truths(text):
    tokens = list(filter(None, text.split(' ')))
    text = ''.join(tokens)
    truths = [
        1. if i == len(token) - 1 else 0.
        for token in tokens
        for i in range(len(token))
    ]
    assert len(text) == len(truths)
    return text, truths


# Observation on bert tokenizer:
# 1. '[UNK]' for oov
# 2. Some tokens may be prefixed with '##'
# 3. Tokens may have longer than length 1 even without '##', e.g. numbers
# 4. Multiple consecutive oov may map to multiple or one '[UNK]' token
def preprocess(text, tokenizer):
    """Convert raw training / testing data to bert tokens format and map their truths.

    >>> preprocess('Ｂ７３７—３００  新世纪  ——  一  １１１１  ＫＫ·Ｄ  。  １２月  ３１日  。  １１００  。  ６—１２  。  Ｄ', create_tokenizer())
    (['[UNK]', '[UNK]', '３０', '##０', '新', '世', '纪', '[UNK]', '[UNK]', '一', '[UNK]', '·', '[UNK]', '。', '１２', '月', '３', '##１', '日', '。', '１１', '##００', '。', '６', '[UNK]', '１２', '。', '[UNK]'], [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 1.0, 0.0, 1.0, 1.0, 0.0, 0.0, 1.0, 1.0, 0.0, 1.0, 0.0, 0.0, 1.0, 1.0, 0.0, 1.0, 1.0, 0.0, 0.0, 1.0, 1.0, 1.0])
    """

    text, truths = _to_text_and_truths(text)
    bert_tokens = tokenizer.tokenize(text)
    bert_truths = []

    j = 0
    i = 0
    while i < len(bert_tokens):
        t = bert_tokens[i][2:] if bert_tokens[i][:2] == '##' else bert_tokens[i]
        l = len(t) if t != '[UNK]' else 1
        if t == text[j:j+l] or t == '[UNK]':
            bert_truths.append(truths[j+l-1])
            j = j + l
            i = i + 1
            continue

        # cannot match, previous token must be '[UNK]'
        assert i > 0 and bert_tokens[i-1] == '[UNK]', (i, text, bert_tokens)
        # assign truth value to the previous '[UNK]' token
        bert_truths[-1] = truths[j]
        j = j + 1


    assert len(bert_tokens) == len(bert_truths)
    return bert_tokens, bert_truths


def postprocess(text, bert_tokens, bert_truths, threshold, seperator='  '):
    """Convert raw training / testing data to bert tokens format and map their truths.

    >>> postprocess('Ｂ７３７—３００  新世纪  ——  一  １１１１  ＫＫ·Ｄ  。  １２月  ３１日  。  １１００  。  ６—１２  。  Ｄ', ['[UNK]', '[UNK]', '３０', '##０', '新', '世', '纪', '[UNK]', '[UNK]', '一', '[UNK]', '·', '[UNK]', '。', '１２', '月', '３', '##１', '日', '。', '１１', '##００', '。', '６', '[UNK]', '１２', '。', '[UNK]'], [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 1.0, 0.0, 1.0, 1.0, 0.0, 0.0, 1.0, 1.0, 0.0, 1.0, 0.0, 0.0, 1.0, 1.0, 0.0, 1.0, 1.0, 0.0, 0.0, 1.0, 1.0, 1.0], 0.5)
    'Ｂ７３７—３００  新世纪  ——  一  １１１１ＫＫ·Ｄ  。  １２月  ３１日  。  １１００  。  ６—１２  。  Ｄ'
    """
    assert len(bert_tokens) == len(bert_truths)
    text, truths = _to_text_and_truths(text)

    truths = []
    j = 0
    i = 0
    while i < len(bert_tokens):
        t = bert_tokens[i][2:] if bert_tokens[i][:2] == '##' else bert_tokens[i]
        l = len(t) if t != '[UNK]' else 1
        if t == text[j:j+l] or t == '[UNK]':
            for k in range(l-1):
                truths.append(0.)
            truths.append(bert_truths[i])
            j = j + l
            i = i + 1
            continue

        # cannot match, previous token must be '[UNK]'
        assert i > 1 and bert_tokens[i-1] == '[UNK]', (i, j, text[j-2:j+3], bert_tokens[i-2:i+3])
        # Assign truth value of '[UNK]' to only the last matching char
        truths[-1] = 0.
        truths.append(bert_truths[i-1])
        j = j + 1

    assert len(truths) == len(text), (len(truths), len(text), text, bert_truths, bert_tokens)

    tokens = []
    for is_token_end, char in zip(truths, text):
        tokens.append(char)
        if is_token_end >= threshold:
            tokens.append(seperator)

    return ''.join(tokens).rstrip(seperator)


def _create_byte_feature(values):
  return tf.train.Feature(bytes_list=tf.train.BytesList(value=[values]))

def _create_int_feature(values):
  feature = tf.train.Feature(int64_list=tf.train.Int64List(value=list(values)))
  return feature


def _create_float_feature(values):
  feature = tf.train.Feature(float_list=tf.train.FloatList(value=list(values)))
  return feature


def text_to_bert_inputs(text, max_seq_length, tokenizer):
  tokens_a, tokens_a_truth = preprocess(text, tokenizer)
  assert len(tokens_a) == len(tokens_a_truth)
  bert_tokens_len = len(tokens_a)
  tokens_b = None

  # Account for [CLS] and [SEP] with "- 2"
  if len(tokens_a) > max_seq_length - 2:
      tokens_a = tokens_a[0:(max_seq_length - 2)]

  # The convention in BERT is:
  # (a) For sequence pairs:
  #  tokens:   [CLS] is this jack ##son ##ville ? [SEP] no it is not . [SEP]
  #  type_ids: 0     0  0    0    0     0       0 0     1  1  1  1   1 1
  # (b) For single sequences:
  #  tokens:   [CLS] the dog is hairy . [SEP]
  #  type_ids: 0     0   0   0  0     0 0
  #
  # Where "type_ids" are used to indicate whether this is the first
  # sequence or the second sequence. The embedding vectors for `type=0` and
  # `type=1` were learned during pre-training and are added to the wordpiece
  # embedding vector (and position vector). This is not *strictly* necessary
  # since the [SEP] token unambiguously separates the sequences, but it makes
  # it easier for the model to learn the concept of sequences.
  #
  # For classification tasks, the first vector (corresponding to [CLS]) is
  # used as as the "sentence vector". Note that this only makes sense because
  # the entire model is fine-tuned.
  tokens = []
  segment_ids = []
  truths = []

  tokens.append("[CLS]")
  segment_ids.append(0)
  truths.append(1.)

  for token, truth in zip(tokens_a, tokens_a_truth):
    tokens.append(token)
    segment_ids.append(0)
    truths.append(truth)

  tokens.append("[SEP]")
  segment_ids.append(0)
  truths.append(1.)

  input_ids = tokenizer.convert_tokens_to_ids(tokens)

  # The mask has 1 for real tokens and 0 for padding tokens. Only real
  # tokens are attended to.
  input_mask = [1] * len(input_ids)

  # Zero-pad up to the sequence length.
  while len(input_ids) < max_seq_length:
    input_ids.append(0)
    input_mask.append(0)
    segment_ids.append(0)
    truths.append(0.)

  assert len(input_ids) == max_seq_length
  assert len(input_mask) == max_seq_length
  assert len(segment_ids) == max_seq_length
  assert len(truths) == max_seq_length

  return input_ids, input_mask, segment_ids, truths


def bert_inputs_to_tfexample(input_ids, input_mask, segment_ids, truths):
  features = collections.OrderedDict()
  # features['text'] = _create_byte_feature(text.encode('utf-8'))
  # features['bert_tokens_len'] = _create_int_feature([bert_tokens_len])
  features["input_ids"] = _create_int_feature(input_ids)
  features["input_mask"] = _create_int_feature(input_mask)
  features["segment_ids"] = _create_int_feature(segment_ids)
  features["truths"] = _create_float_feature(truths)

  return tf.train.Example(features=tf.train.Features(feature=features))


def text_to_tfexample(text, max_seq_length, tokenizer):
    return bert_inputs_to_tfexample(*text_to_bert_inputs(text, max_seq_length, tokenizer))


def feature_spec(seq_length):
    return {
        # "text": tf.VarLenFeature(tf.string),
        # "bert_tokens_len": tf.FixedLenFeature([1], tf.int64),
        "input_ids": tf.FixedLenFeature([seq_length], tf.int64),
        "input_mask": tf.FixedLenFeature([seq_length], tf.int64),
        "segment_ids": tf.FixedLenFeature([seq_length], tf.int64),
        "truths": tf.FixedLenFeature([seq_length], tf.float32),
    }


# def model_prediction_to_result(predictions):
#     text = predictions['text'].eval().values[0].decode('utf-8')
#     bert_tokens_len = features['bert_tokens_len'].eval()[0]
#     bert_tokens = tokenizer.convert_ids_to_tokens(features['input_ids'].eval()[1:bert_tokens_len+1])
#     bert_truths = features['truths'].eval()[1:bert_tokens_len+1]
#
#     print("Final Result:", postprocess(text, bert_tokens, bert_truths, 0.5))



if __name__ == "__main__":
    pass



import tensorflow as tf
from transform import feature_spec, preprocess, text_to_bert_inputs

def _deserialize(serialized, name_to_features):
  """Decodes a record to a TensorFlow example."""
  example = tf.parse_single_example(serialized, name_to_features)

  # tf.Example only supports tf.int64, but the TPU only supports tf.int32.
  # So cast all int64 to int32.
  for name in list(example.keys()):
    t = example[name]
    if t.dtype == tf.int64:
      t = tf.to_int32(t)
    example[name] = t

  return example


def input_fn_builder(input_file, seq_length, shuffle, repeat, drop_remainder):
  """Creates an `input_fn` closure to be passed to TPUEstimator."""

  def input_fn(params):
    """The actual input function."""
    batch_size = params["batch_size"]

    # For training, we want a lot of parallel reading and shuffling.
    # For eval, we want no shuffling and parallel reading doesn't matter.
    d = tf.data.TFRecordDataset(input_file)
    if shuffle:
      d = d.shuffle(buffer_size=1024)
    if repeat:
      d = d.repeat()

    d = d.apply(
        tf.data.experimental.map_and_batch(
            lambda serialized: _deserialize(serialized, feature_spec(seq_length)),
            batch_size=batch_size,
            drop_remainder=drop_remainder))
    return d

  return input_fn


def serving_input_fn_builder(seq_length, batch_size):
    def serving_input_receiver_fn():
      """An input receiver that expects a serialized tf.Example."""
      serialized_tf_example = tf.placeholder(dtype=tf.string,
                                             shape=[batch_size],
                                             name='input_example_tensor')
      receiver_tensors = {'examples': serialized_tf_example}
      features = tf.parse_example(serialized_tf_example, feature_spec(seq_length))
      return tf.estimator.export.ServingInputReceiver(features, receiver_tensors)

    return serving_input_receiver_fn


# TODO: padding insteads of drop_remainder
def predict_input_fn_builder(bert_inputs, seq_length, tokenizer, drop_remainder=False):
  all_input_ids = []
  all_input_mask = []
  all_segment_ids = []
  all_truths = []

  for input_ids, input_mask, segment_ids, truths in bert_inputs:
      all_input_ids.append(input_ids)
      all_input_mask.append(input_mask)
      all_segment_ids.append(segment_ids)
      all_truths.append(truths)

  num_examples = len(bert_inputs)

  def input_fn(params):
      d = tf.data.Dataset.from_tensor_slices({
          "input_ids":
              tf.constant(
                  all_input_ids, shape=[num_examples, seq_length],
                  dtype=tf.int32),
          "input_mask":
              tf.constant(
                  all_input_mask,
                  shape=[num_examples, seq_length],
                  dtype=tf.int32),
          "segment_ids":
              tf.constant(
                  all_segment_ids,
                  shape=[num_examples, seq_length],
                  dtype=tf.int32),
          "truths":
              tf.constant(
                  all_truths,
                  shape=[num_examples, seq_length],
                  dtype=tf.float32),
      })
      d = d.batch(batch_size=params['batch_size'], drop_remainder=drop_remainder)
      return d

  return input_fn

def serving_input_fn_builder(seq_length):
   return tf.estimator.export.build_raw_serving_input_receiver_fn({
    'input_ids': tf.placeholder(
        shape=[None, seq_length],
        dtype=tf.int32,
        name='input_ids_ph'
    ),
    'input_mask': tf.placeholder(
        shape=[None, seq_length],
        dtype=tf.int32,
        name='input_mask_ph'
    ),
    'segment_ids': tf.placeholder(
        shape=[None, seq_length],
        dtype=tf.int32,
        name='segment_ids_ph'
    ),
    'truths': tf.placeholder(
        shape=[None, seq_length],
        dtype=tf.float32,
        name='truths_ph'
    )
  })



import os
import tensorflow as tf
from input import input_fn_builder, serving_input_fn_builder
from tqdm import tqdm
import sys

flags = tf.flags
FLAGS = flags.FLAGS

flags.DEFINE_string("assets_dir", "assets", "The assets directory generated by assets.py.")
flags.DEFINE_string("checkpoint_dir", 'ckpt', "The directory for storing model check points.")
flags.DEFINE_string("gs_bert_model_ch_dir", 'gs://berserker/repo/assets/chinese_L-12_H-768_A-12', "A google storage path to unzipped BERT chinese_L-12_H-768_A-12 model.")
flags.DEFINE_string("train_file", "dataset/train_128.tfrecords", "The training file output by dataset.py.")

flags.DEFINE_integer("max_seq_length", 512, "Maximum sequence length.")
flags.DEFINE_integer("batch_size", 64, "The training, validation and prediction batch size.")

flags.DEFINE_bool("do_train", False, "Train the model.")
flags.DEFINE_float("learning_rate", 2e-5, "The learning rate.")
flags.DEFINE_integer("train_steps", 15243//64, "Number of training steps.")
flags.DEFINE_float("warmup_proportion", 0.1, "")
flags.DEFINE_integer("save_checkpoints_steps", 15243//64, "Number of steps to save a checkpoint.")


flags.DEFINE_bool("do_eval", False, "Evaluate the model.")
flags.DEFINE_string("eval_file", "dataset/val_128.tfrecords", "The validation file output by dataset.py.")
flags.DEFINE_integer("eval_steps", 3811//64, "Number of validation steps.")


flags.DEFINE_bool("do_predict", False, "Make prediction.")
flags.DEFINE_string("predict_file", "assets/icwb2-data/pku_test.utf8", "The input file to be tokenized.")
flags.DEFINE_string("predict_output", "pku_pred.utf8", "The output file for tokenized result.")


flags.DEFINE_bool("do_export", False, "Export model.")
flags.DEFINE_string("output_dir", 'gs://berserker/export', "Output path for exported model.")


flags.DEFINE_bool("use_tpu", False, "Use TPU for training.")
flags.DEFINE_integer("num_tpu_cores", 8, "The number of TPU cores.")
flags.DEFINE_string("tpu_name", None, "TPU worker.")

def main(_):
    tf.logging.set_verbosity(tf.logging.INFO)
    # tf.gfile.MakeDirs(FLAGS.output_dir)

    sys.path += [os.path.join(FLAGS.assets_dir, 'bert')]
    from model import model_fn_builder
    import modeling

    model_fn = model_fn_builder(
        bert_config=modeling.BertConfig.from_json_file(
            os.path.join(FLAGS.gs_bert_model_ch_dir, 'bert_config.json')
        ),
        init_checkpoint=os.path.join(FLAGS.gs_bert_model_ch_dir, 'bert_model.ckpt'),
        use_tpu=FLAGS.use_tpu,
        use_one_hot_embeddings=True if FLAGS.use_tpu else False,
        learning_rate=FLAGS.learning_rate,
        num_train_steps=FLAGS.train_steps,
        num_warmup_steps=int(FLAGS.train_steps * FLAGS.warmup_proportion)
    )

    run_config = tf.contrib.tpu.RunConfig(
        cluster=tf.contrib.cluster_resolver.TPUClusterResolver(FLAGS.tpu_name) if FLAGS.use_tpu else None,
        model_dir=FLAGS.checkpoint_dir,
        save_checkpoints_steps=FLAGS.train_steps,
        tpu_config=tf.contrib.tpu.TPUConfig(
            iterations_per_loop=FLAGS.train_steps,
            num_shards=FLAGS.num_tpu_cores,
            per_host_input_for_training=tf.contrib.tpu.InputPipelineConfig.PER_HOST_V2
        )
    )

    estimator = tf.contrib.tpu.TPUEstimator(
        use_tpu=FLAGS.use_tpu,
        model_fn=model_fn,
        config=run_config,
        train_batch_size=FLAGS.batch_size,
        eval_batch_size=FLAGS.batch_size,
        predict_batch_size=FLAGS.batch_size
    )

    tf.logging.info('Setup success...')

    if FLAGS.do_train:
        estimator.train(
            input_fn=input_fn_builder(
                input_file=FLAGS.train_file,
                seq_length=FLAGS.max_seq_length,
                shuffle=True,
                repeat=True,
                drop_remainder=FLAGS.use_tpu
            ),
            steps=FLAGS.train_steps,
        )

    if FLAGS.do_eval:
        estimator.evaluate(
            input_fn=input_fn_builder(
                input_file=FLAGS.eval_file,
                seq_length=FLAGS.max_seq_length,
                shuffle=False,
                repeat=False,
                drop_remainder=FLAGS.use_tpu
            ),
            steps=FLAGS.eval_steps,
        )

    if FLAGS.do_predict:
        from transform import create_tokenizer, text_to_bert_inputs, postprocess, preprocess
        from input import predict_input_fn_builder
        import numpy as np
        tokenizer = create_tokenizer(
            os.path.join(FLAGS.assets_dir, 'chinese_L-12_H-768_A-12', 'vocab.txt')
        )

        import pandas as pd
        texts = pd.read_csv(FLAGS.predict_file, header=None, sep='^')[0]

        # texts = [
        #     '共同创造美好的新世纪——二○○一年新年贺词',
        #     '（二○○○年十二月三十一日）（附图片1张）',
        #     '女士们，先生们，同志们，朋友们：',
        #     '2001年新年钟声即将敲响。人类社会前进的航船就要驶入21世纪的新航程。中国人民进入了向现代化建设第三步战略目标迈进的新征程。',
        #     '在这个激动人心的时刻，我很高兴通过中国国际广播电台、中央人民广播电台和中央电视台，向全国各族人民，向香港特别行政区同胞、澳门特别行政区同胞和台湾同胞、海外侨胞，向世界各国的朋友们，致以新世纪第一个新年的祝贺！',
        #     '过去的一年，是我国社会主义改革开放和现代化建设进程中具有标志意义的一年。在中国共产党的领导下，全国各族人民团结奋斗，国民经济继续保持较快的发展势头，经济结构的战略性调整顺利部署实施。西部大开发取得良好开端。精神文明建设和民主法制建设进一步加强。我们在过去几年取得成绩的基础上，胜利完成了第九个五年计划。我国已进入了全面建设小康社会，加快社会主义现代化建设的新的发展阶段。',
        #     '面对新世纪，世界各国人民的共同愿望是：继续发展人类以往创造的一切文明成果，克服20世纪困扰着人类的战争和贫困问题，推进和平与发展的崇高事业，创造一个美好的世界。',
        #     '我们希望，新世纪成为各国人民共享和平的世纪。在20世纪里，世界饱受各种战争和冲突的苦难。时至今日，仍有不少国家和地区的人民还在忍受战火的煎熬。中国人民真诚地祝愿他们早日过上和平安定的生活。中国人民热爱和平与自由，始终奉行独立自主的和平外交政策，永远站在人类正义事业的一边。我们愿同世界上一切爱好和平的国家和人民一道，为促进世界多极化，建立和平稳定、公正合理的国际政治经济新秩序而努力奋斗。' * 10
        # ]

        bert_inputs = []
        bert_inputs_lens = []
        SEQ_LENGTH = FLAGS.max_seq_length - 2
        for text in texts:
            while len(text) > 0:
                bert_input = text_to_bert_inputs(text[:SEQ_LENGTH], FLAGS.max_seq_length, tokenizer)
                bert_inputs_lens.append(len(preprocess(text[:SEQ_LENGTH], tokenizer)[0]))
                bert_inputs.append(bert_input)
                text = text[SEQ_LENGTH:]

        # Pad to prediction batch size
        while len(bert_inputs) % FLAGS.batch_size != 0:
            bert_input = text_to_bert_inputs('', FLAGS.max_seq_length, tokenizer)
            bert_inputs_lens.append(len(preprocess('', tokenizer)[0]))
            bert_inputs.append(bert_input)

        tf.logging.info(len(bert_inputs))

        results = estimator.predict(
            input_fn=predict_input_fn_builder(
                bert_inputs=bert_inputs,
                seq_length=FLAGS.max_seq_length,
                tokenizer=tokenizer,
                drop_remainder=FLAGS.use_tpu
            )
        )

        results = [result for result in results]

        for threshold in np.linspace(0.1, 0.9, 9):
            results_itr = iter(results)
            bert_inputs_itr = iter(bert_inputs)
            bert_inputs_lens_itr = iter(bert_inputs_lens)

            with open('%.1f_'%threshold + FLAGS.predict_output, 'w') as f:
                for text in texts:
                    original_text = text
                    prediction = np.array([])
                    bert_tokens = []

                    while len(text) > 0:
                        (input_ids, _, _, _) = next(bert_inputs_itr)
                        bert_inputs_len = next(bert_inputs_lens_itr)
                        result = next(results_itr)
                        prediction = np.concatenate((prediction, result['predictions'][1:1+bert_inputs_len]))
                        bert_tokens += tokenizer.convert_ids_to_tokens(input_ids[1:1+bert_inputs_len])
                        text = text[SEQ_LENGTH:]

                    tokenized_text = postprocess(original_text, bert_tokens, prediction, threshold=threshold)
                    print(tokenized_text, file=f)
                    # tf.logging.info(" Input: %s"%original_text)
                    # tf.logging.info("Output: %s"%tokenized_text)


    if FLAGS.do_export:
        estimator._export_to_tpu = False
        estimator.export_savedmodel(
            export_dir_base=FLAGS.output_dir,
            serving_input_receiver_fn=serving_input_fn_builder(
                seq_length=FLAGS.max_seq_length
            ),
            strip_default_attrs=True
        )

if __name__ == "__main__":
  tf.app.run()
